/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2011 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

 Part of the example are taken from the QuantLib, 
 file test-suite/swingoption.cpp 

 compile:
export LD_LIBRARY_PATH=/usr/lib/R/lib:/usr/lib/R/library/RInside/lib:/usr/lib/R/library/Rcpp/lib:/usr/lib/R/library/RInside/lib:/usr/lib/R/library/Rcpp/lib


g++ -I/usr/local/include  -I/usr/lib/R/include -I/usr/lib/R/library/RInside/include -I/usr/lib/R/library/Rcpp/include -I/home/spanderen/workspace/QuantLib -I/usr/local/include/boost-1_38 -O3 -o fwdswing fwdswing.cpp -L/home/spanderen/workspace/QuantLib/ql/.libs -lQuantLib  -L/usr/lib/R/library/RInside/lib -lRInside -L/usr/lib/R/library/Rcpp/lib -lRcpp -L/usr/lib/R/lib -lR -lglpk -lboost_thread-gcc42-mt
*/


#include <ql/math/functional.hpp>
#include <ql/quotes/simplequote.hpp>
#include <ql/time/daycounters/actualactual.hpp>
#include <ql/instruments/vanillaswingoption.hpp>
#include <ql/math/randomnumbers/rngtraits.hpp>
#include <ql/math/statistics/generalstatistics.hpp>
#include <ql/termstructures/yield/zerocurve.hpp>
#include <ql/termstructures/yield/flatforward.hpp>
#include <ql/methods/montecarlo/multipathgenerator.hpp>
#include <ql/experimental/processes/extouwithjumpsprocess.hpp>
#include <ql/experimental/processes/extendedornsteinuhlenbeckprocess.hpp>
#include <ql/experimental/finitedifferences/fdsimpleextoujumpswingengine.hpp>
#include <ql/experimental/finitedifferences/fdblackscholesvanillaengine.hpp>

#include <boost/shared_array.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <glpk.h>
#include <RInside.h>  // for the embedded R via RInside

#include <fstream>

#ifndef NTHREADS
#define NTHREADS 1
#endif


using namespace QuantLib;

namespace {
    boost::shared_ptr<YieldTermStructure>
    flatRate(const Date& today, Rate forward, const DayCounter& dc) {
        return boost::shared_ptr<YieldTermStructure>(
            new FlatForward(today, Handle<Quote>(
                   boost::shared_ptr<Quote>(new SimpleQuote(forward))), dc));
    }

    class ForwardPayoff : public StrikedTypePayoff {
    public:
	    ForwardPayoff(Real strike) 
		: StrikedTypePayoff(Option::Call, strike) { }

	
	    std::string name() const { return  "vanilla"; }
	    std::string description() const {
		    std::ostringstream result;
			result << name() << " " << "fwd";
			return result.str();
		}

	    Real operator()(Real price) const {
            return 0.0, price - strike();
		}
    };

    const Real beta = 200.0;
    const Real eta  = 1.0/0.4;
    const Real jumpIntensity = 4.0;
    const Real speed = 7.0;
    const Real volatility = 1.4;
    const Rate irRate = 0.00;

    class Worker {
    public:
        typedef FdmExtOUJumpModelInnerValue::Shape Shape;

        Worker(Size minEx, Size maxEx, 
               Real strike,
               const std::vector<std::pair<Time, Real> >& fwdCurve,
               const std::vector<std::pair<Date, Size> >& swingDateTimes)
        : minExerciseRights_(minEx),
          maxExerciseRights_(maxEx),
          strike_(strike),
          shape_(boost::shared_ptr<Shape>(new Shape(fwdCurve.size()))), 
          swingDateTimes_(swingDateTimes),  
          data(new Real[3]) { 

            for (Size i=0; i < fwdCurve.size(); ++i) {
                const Time t = fwdCurve[i].first;
                const Real f = std::log(fwdCurve[i].second)
                    -square<Real>()(volatility)/(4*speed)*(1-std::exp(-2*speed*t))
                    -jumpIntensity/beta*std::log((1-std::exp(-beta*t)/eta)/(1-1/eta));
                (*shape_)[i] = std::pair<Time, Real>(t,f);
            }
        }

        void operator()() {
            boost::mutex::scoped_lock lock(mutex[Size(strike_/10) % NTHREADS]);

            DayCounter dayCounter = ActualActual();
            const Date settlementDate = Settings::instance().evaluationDate();
            boost::shared_ptr<StrikedTypePayoff> payoff(new ForwardPayoff(strike_));

            std::vector<Date> exerciseDates(swingDateTimes_.size());
            std::vector<Time> exerciseTimes(exerciseDates.size());
            std::vector<Size> exerciseSec(exerciseDates.size());
            for (Size i=0; i < exerciseDates.size(); ++i) {
                exerciseDates[i] = swingDateTimes_[i].first;
                exerciseTimes[i] 
                    = dayCounter.yearFraction(settlementDate, exerciseDates[i])
                       + swingDateTimes_[i].second/(24*365*3600.);
                exerciseSec[i] = swingDateTimes_[i].second;
            }

            boost::shared_ptr<SwingExercise> swingExercise(
                                 new SwingExercise(exerciseDates, exerciseSec));

            TimeGrid grid(exerciseTimes.begin(), exerciseTimes.end(), 1);
            std::vector<Size> exerciseIndex(exerciseDates.size());
            for (Size i=0; i < exerciseIndex.size(); ++i) {
                exerciseIndex[i] = grid.closestIndex(exerciseTimes[i]);
            }

            Array x0(2);
            x0[0] = 0.0; x0[1] = 0.0;

            boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> ouProcess(
               new ExtendedOrnsteinUhlenbeckProcess(speed, volatility, x0[0],
                                                 constant<Real, Real>(x0[0])));
            boost::shared_ptr<ExtOUWithJumpsProcess> jumpProcess(
                new ExtOUWithJumpsProcess(ouProcess, x0[1], beta, 
                                          jumpIntensity, eta));

            boost::shared_ptr<YieldTermStructure> rTS(
                                flatRate(settlementDate, irRate, dayCounter));

            const Size nrTrails = 5000;
            typedef PseudoRandom::rsg_type rsg_type;
            typedef MultiPathGenerator<rsg_type>::sample_type sample_type;
            rsg_type rsg = PseudoRandom::make_sequence_generator(
                    jumpProcess->factors()*(grid.size()-1), BigNatural(421));

            MultiPathGenerator<rsg_type> generator(jumpProcess, grid, rsg, false);

            GeneralStatistics npv;
                
            // set-up linear optimization problem
            {
#if NTHREADS > 1
                boost::mutex::scoped_lock glp_lock(mutex[NTHREADS]);
#endif
                glp_prob *lp;
                lp = glp_create_prob();
            
                glp_set_prob_name(lp, "swingoption");
                glp_set_obj_dir(lp, GLP_MAX);
            
                glp_add_rows(lp, 1);
                glp_set_row_name(lp, 1, "exerciseRights");
                glp_set_row_bnds(lp, 1, GLP_DB, minExerciseRights_, maxExerciseRights_);
            
                boost::scoped_array<int> ia(new int[1+exerciseTimes.size()]);
                boost::scoped_array<int> ja(new int[1+exerciseTimes.size()]);
                boost::scoped_array<Real> ar(new double[1+exerciseTimes.size()]);
            
                glp_add_cols(lp, exerciseTimes.size());
                for (Size i=1; i <= exerciseTimes.size(); ++i) {
                    char varName[16]; 
                    sprintf(varName,"x%i",i);
                    
                    glp_set_col_name(lp, i, varName);
                    glp_set_col_bnds(lp, i, GLP_DB, 0.0, 1.0);
                    
                    ja[i] = Real(i);
                    ia[i] = ar[i] = 1.0;
                }
                glp_load_matrix(lp, exerciseTimes.size(), ia.get(), ja.get(), ar.get());
                
                glp_smcp simplexParam;
                glp_init_smcp(&simplexParam);
                simplexParam.msg_lev = GLP_MSG_ERR;
                
                // use MC plus perfect forecast to find an upper bound
                for (Size n=0; n < nrTrails; ++n) {
                    sample_type path = generator.next();
                    std::vector<Real> exerciseValues(exerciseTimes.size());
                    for (Size k=0; k < exerciseTimes.size(); ++k) {
                        const Time t = exerciseTimes[k];
                        const Real x = path.value[0][exerciseIndex[k]];
                        const Real y = path.value[1][exerciseIndex[k]];
                        
                        const Real z = std::lower_bound(
                            shape_->begin(), shape_->end(),
                            std::pair<Time, Real>(t-std::sqrt(QL_EPSILON), 
                                                  0.0))->second;
                        
                        const Real s = std::exp(z + (x+y));
                    
                        exerciseValues[k] 
                            =(*payoff)(s)*rTS->discount(exerciseDates[k]);
                        glp_set_obj_coef(lp, k+1, exerciseValues[k]);
                    }
                    glp_simplex(lp, &simplexParam);
                    QL_REQUIRE(glp_get_status(lp) == GLP_OPT 
                               || glp_get_status(lp) == GLP_FEAS, 
                               "infeasable problem");
                    npv.add(glp_get_obj_val(lp));
                }
                glp_delete_prob(lp);
            }

            data[0] = npv.mean();
            data[1] = npv.errorEstimate();
            
            VanillaSwingOption swingOption(payoff, swingExercise,
                                           minExerciseRights_, maxExerciseRights_);
            
            boost::shared_ptr<PricingEngine> swingEngine(
               new FdSimpleExtOUJumpSwingEngine(jumpProcess, rTS, 1, 50,10, shape_));
            swingOption.setPricingEngine(swingEngine);
            data[2] = swingOption.NPV();
        }

        Real getMCNpv()   const { return data[0]; }
        Real getMCError() const { return data[1]; }        
        Real getFDMNpv()  const { return data[2]; }
        
    private:
        const Size minExerciseRights_;
        const Size maxExerciseRights_;
        const Real strike_;
        const boost::shared_ptr<Shape> shape_;
        const std::vector<std::pair<Date, Size> > swingDateTimes_;
        const boost::shared_array<Real> data;

        static boost::shared_array<boost::mutex> mutex;
    };

    boost::shared_array<boost::mutex> Worker::mutex(new boost::mutex[NTHREADS+1]);
}

int main(int argc, char *argv[]) {
    Date settlementDate = Date(28, Sep, 2010);
    Settings::instance().evaluationDate() = settlementDate;

    const Size maturityInWeeks = 12;
    const Size maturityInDays  = 7*maturityInWeeks;
    const Size maturityInHours = 24*maturityInDays;
    Date maturityDate = settlementDate + Period(maturityInWeeks, Weeks);

    std::vector<std::pair<Date, Size> > swingDateTimes;
    swingDateTimes.reserve(maturityInHours);

    std::vector<std::pair<Time, Real> > fwdCurve;
    std::vector<Real> prices;
    fwdCurve.reserve(maturityInHours);
    prices.reserve(maturityInHours);

    // read shape from file
    // format is Date, 0h price, 1h price, 2h price, ... 23h price
    // example:4-Oct-2010 36.73 32.09 27.32 23.22 25.71 35.60 47.32 ...
    std::ifstream fwd("EEX_2010-2013.txt");
    QL_REQUIRE(fwd.is_open(), "can't find shape file EEX_2010-2013.txt");
    
    for (Size i=0; i < maturityInDays; ++i) {
        std::string str;
        fwd >> str;

        Date date = settlementDate+Period(i, Days);
        for (Size j=0; j < 24; ++j) {
            swingDateTimes.push_back(std::pair<Date, Size>(date, j*3600));

            Real p;
            fwd >> p;
            prices.push_back(p);
            fwdCurve.push_back(std::pair<Time, Real>((i + j/24.)/365., p));
        }
    }
    fwd.close();
 
    const Size minExerciseRights = 100;
    const Size maxExerciseRights = 500;

    const Real strikes[] = { 40, 50, 60, 70, 80, 90, 100, 110, 120 };
    const Size n=sizeof(strikes)/sizeof(Real);

    std::vector<Real> fdmNPV, mcNPV, mcError;
    std::vector<boost::shared_ptr<Worker> > workers(n);
    std::vector<boost::shared_ptr<boost::thread> > workerThreads(n);

    for (Size i=0; i < n; ++i) {
        workers[i] = boost::shared_ptr<Worker>(
            new Worker(minExerciseRights, maxExerciseRights,
                       strikes[i], fwdCurve, swingDateTimes));
        workerThreads[i] = boost::shared_ptr<boost::thread>(
                                    new boost::thread(*workers[i].get()));
    }

    for (Size i=0; i < n; ++i) {
        workerThreads[i]->join();
        fdmNPV.push_back(workers[i]->getFDMNpv());
        mcNPV.push_back(workers[i]->getMCNpv());
        mcError.push_back(workers[i]->getMCError());
        printf("%f %f %f %f\n",strikes[i], fdmNPV.back(), 
               mcNPV.back(), mcError.back());
    }

    // embedded R script. Remove the following lines if you don't use RInside
    RInside R(argc, argv);

    R["x"] = fdmNPV;
    R["y"] = mcNPV;
    R["u"] = mcError;
    R["p"] = prices;
    R["strikes"] = std::vector<Real>(strikes, strikes + n);
    R["minEx"] = minExerciseRights;
    R["maxEx"] = maxExerciseRights;

    std::string rScript[] = {

        "if (nchar(installed.packages()[,'Package']['Cairo'])==5) { "
        "  library(\"Cairo\"); ",
        "  if (Cairo.capabilities()['png']) { "
        "    CairoPNG(\"plot.png\") "
        "  } else { "
        "  pdf(\"plot.pdf\") }"
        "} else { pdf(\"plot.pdf\") }", 

        "library(\"Hmisc\")",
        "library(\"foreach\")",

        "f <- function(payoff, minEx, maxEx) { "
        "   v <- payoff[(minEx+1):maxEx]; sum(payoff[1:minEx])+sum(v[v>0])"
        "} ",

        "par(mar=c(4,4,2,2)+0.1)",

        "errbar(strikes,y,y-u,y+u,xlab=\"Strike (in EUR/MWh)\","
        "    ylab=\"Price (in EUR)\", ylim=c(-5000,25000),pch=16, type=\"b\",lty=1)",

        "points(strikes, x, pch=3, type=\"b\",lty=1)",

        "points(strikes, foreach(i = strikes) %do% "
        "   f(sort(p-i,decreasing=TRUE),minEx,maxEx),pch=22, type=\"b\",lty=1)",

        "legend(x=70,y=24000,legend=c(\"Intrinsic Value\","
        "                             \"Linear Programming (MC)\","
        "                             \"Dynamic Programming (FDM)\"), pch=c(22,16,3))"
    };


    for (Size i=0; i < sizeof(rScript)/sizeof(std::string); ++i) {
        R.parseEvalQ(rScript[i]);
    }

    return 0;
}
